//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.


#include <iostream>
#include <cxxtls/foreach.h>
#include "BrashCommands.h"
#include <cxxtls/stringexpression.h>
#include <stdlib.h>
#include <fstream>
#include <exception>
#include <time.h>
#include <signal.h>

#include <cxxtls/thread.h>

using namespace cxxtls;
using namespace std;


static void setBrashUserHome(StringExpr::ExecutionVariables &env);   // set variable BRASH_USER_HOME

string interpreterExecutableName;

BrashScriptCommandHandler *brashGlobalScriptCommandHandler = 0;  // should be passed to all execute() methods by the handle() function when commands are invoked
                                                                 // but for the moment, I'm going to leave it here.

static void   SetStringExpressionUndefinedVariablesHandlers();

bool noReadBrashrc = false; // used by ReadConsole

string specifiedRCFileName;  // used in brashCommands.cxx

static void printCommandLineOptions()
{
   cout << "\r\n\r\nBrash version " BRASH_VERSION_STRING 
   #ifdef _MSC_VER
           " for Microsoft Windows."
   #else
           " for Linux/Unix."
   #endif
            "\r\n\r\n"
            "Usage:\r\n\r\n"
            "brash [options]\r\n"
            "\r\n"
            "      -h                            print this help message\r\n"
            "      --help                        print this help message\r\n"
            "      -v                            print brash version information\r\n"
            "      -c cmd                        execute a single command line parameter as a command\r\n"
            "      -script file [arg ...]        execute a file and pass command line arguments to it\r\n"
            "      -threads count                set number of running threads (default=20)\r\n"
            "      -debugLevel number            set diagnostic log level.  (default=0, max=3)\r\n"
            "      -norc                         suppress reading of $HOME/.brashrc\r\n"
            "      -rc file                      use file instead of .brashrc\r\n"
            "\r\n"
            "Brash is a multi-threaded implementation of the bourne shell.  It does not rely on fork/exec\r\n"
            "Some bash features are included as well -- though compliance is incomplete.\r\n"
            "\r\n"
            "In Brash/Bourne, the final stage of a pipe runs in the current shell, not in a sub-shell.\r\n"
            "This means that you can put a read statement in the final stage and expect it to populate\r\n"
            "variables in the current shell, but in Brash the final state is in the current shell.\r\n"
            "\r\n"
            "In this version, if-statements and while-statements only permit one command\r\n"
            "to define the condition, unlike bourne and brash which allow multiple commands\r\n"
            "separated by semicolons.  You can, of course, use {} or () grouping in brash to\r\n"
            "achieve the same effect.\r\n"
            "\r\n"
            "If you do not specify the -norc option, Brash will read $BRASH_USER_HOME/.brashc\r\n"
            "before executing any other commands you might have given it.\r\n"
            "\r\n"
            "Brash defines the following variables:\r\n"
            "    BRASH_USER_HOME       -- either $HOME on unix, or %USERPROFIEL% on Windows\r\n"
            "    BRASH_SHELL           -- the name of this executable\r\n"
            "    BRASH_VERSION         -- the current version of this executable\r\n"
            "    BRASH_PATH_SEPARATOR  -- ';' or ':' depending on the os\r\n"
            "    BRASH_DIR_SEPARATOR   -- / or \\ depending on the os\r\n"
            "    BRASH_SYSROOT         -- either C:\\ or / depending on the os\r\n"
            "    BRASH_OS              -- Either MSWindows or unix depding on the os\r\n"
            "    BRASH_USER            -- Either %USERNAME% or your unix user id\r\n"
            "    BRASH_HOSTNAME        -- Either %COMPUTERNAME or your unix hostname \r\n"
            "    BRASH_NO_BANNER       -- suppress 'welcome screen' at the console\r\n"
            "\r\n"
            "Debug Levels are as follows:\r\n"
            "   0 -- normal:     major errors only\r\n"
            "   1 -- diagnostic: unusual situations\r\n" 
            "   2 -- debug:      info needed to diagnose program bugs\r\n"
            "   3 -- heavy:      puke my guts out -- I mean really!\r\n"
            "\r\n"
            "Console Window Edit Keys:\r\n"
            "   SCREEN_RESIZED -- screen resizes are interpreted as keystrokes\r\n"
            "   ^C             -- user requests that everything stop\r\n"
            "   HOM            -- go to the beginning of the current input line\r\n"
            "   ^A             -- ditto\r\n"
            "   END            -- go to the end of the current input line\r\n"
            "   ^E             -- ditto\r\n"
            "   RightArrow     -- move one position to the right in the current line\r\n"
            "   LeftArrow      -- move one position to th eleft in the current line\r\n"
            "   ^K             -- delete the rest of the line\r\n"
            "   ^W             -- move the cursor to the next word in the current line\r\n"
            "   ^Q             -- move the cursor to the previous word in the current line\r\n"
            "   UpArrow        -- display the previous line from the command history\r\n"
            "   DownArrow      -- display the next line from the command history\r\n"
            "   Backspace      -- delete the previous character\r\n"
            "   Delete         -- delete the current character\r\n"
            "   TAB (^I)       -- 'tab completion':  find the a file that matches the word\r\n"
            "                     to the left of the cursor\r\n"
        ;
}

int main(int args, char **argv)
{

   CXXTLS_SET_PROGRAM_TRACE_LEVEL(TT_NONE);  // by default, log nothing


   SetStringExpressionUndefinedVariablesHandlers();

try
{



   // throw BrashCommandLine::TooLongException("a parm", 1.3, "another parm", double(4.9));


   interpreterExecutableName = *argv++;
   --args;



//  just a test to make sure it works   handler.push(new exitCommand(&handler,13));

   StringExpr::ExecutionVariables env;


   bool launchConsoleReader = false;
   bool launchFileReader    = false;


   string scriptFileName;

   env.includeProgramEnvironment();

   env.set("IFS", " \t\n\r");  // set the default inter-field-separator which is used in for statements to split variables after they have been
                               // expanded.  It is used in a few other places as well.


   setBrashUserHome(env);

   StringExpr::ExecutionContext::ArgsVector commandLine;  

   // for the moment:

   commandLine.push_back(interpreterExecutableName); // just a debug measure

   bool dash_c_option = false;
   int  threadCount   = 20;

   // handle invocation options

   string commandString;

   while(*argv && argv[0][0] == '-')
   {
       // handle command line options

       if(   argv[0][0] ==  '-'
          && argv[0][1] == 0
         )
       {
          break; // we have hit an option named "-"
       }

       if(   argv[0][0] == '-'
          && argv[0][1] == '-'
          && argv[0][2] == 0
         )
       {
          break; // we have hit an option named "--"
       }


       string option(*argv++);

       if(option == "-rc")
       {
          if(*argv == 0)
          {
              cerr << "Error:  expected a parameter for the -rc option" << endl;
              exit(1);
          }

          else
          if(argv[0][0] == '-')
          {
              cerr << "Error:  '"  << *argv << "' is not a valid option for -rc" << endl;
              exit(1);
          }

          specifiedRCFileName = *argv++;

       }
       else
       if(option == "-norc")
          noReadBrashrc = true;
       else
       if(option == "-h" || option == "--help")
       {
           printCommandLineOptions();
           exit(0);
       }
       else
       if(option == "-v")
       {
          cout << "Brash version " BRASH_VERSION_STRING
#ifdef _MSC_VER
               << " for Microsoft Windows"
#else
               << " for Linux/Unix"
#endif
               << endl;
          exit(0);

       }
       else
       if(option == "-script")
       {
           if(*argv == 0)
           {
               cerr << "Error:  expected filename to follow -script" << endl;
               exit(1);
           }

           scriptFileName = *argv++;
           launchFileReader = true;
       }
       else
       if(option == "-c")
       {
          dash_c_option = true;
       }
       else
       if(option == "-threads")
       {
          if(*argv == 0)
          {
             cerr << "Error:  expected <count> to follow -threads\r\n";
             exit(1);
          }
          int tmp = 0;

          tmp = atoi(*argv++);

          if(tmp > 0)
             threadCount = tmp;
          else
          {
             cerr << "Error:  invalid <count> option to -threads\r\n";
             exit(1);
          }
       }
       else
       if(option == "-debugLevel" )
       {
          if(*argv == 0)
          {
             cerr << "Error:  expected <debugLevel> to follow -debugLevel\r\n";
             cerr << "        debug levels are 0-3:\r\n";
             cerr << "            0:  NONE\r\n";
             cerr << "            1:  NORMAL LOGGING\r\n";
             cerr << "            2:  DEBUG\r\n";
             cerr << "            3:  HEAVY DEBUGGING\r\n";
             cerr <<"Note:  3 is VERY HEAVY\r\n";
             exit(1);
          }
          int tmp = 0;

          tmp = atoi(*argv++);

          if(tmp < 0 || tmp > 3)
          {
             cerr << "Error:  invalid <debugLevel> value to -debugLevel\r\n";
             cerr << "        use a number in the range of 0 to 3:\r\n";
             cerr << "        debug levels are 0-3:\r\n";
             cerr << "            0:  NONE\r\n";
             cerr << "            1:  NORMAL LOGGING\r\n";
             cerr << "            2:  DEBUG\r\n";
             cerr << "            3:  HEAVY DEBUGGING\r\n";
             cerr <<"Note:  3 is VERY HEAVY\r\n";
             exit(1);
          }

              CXXTLS_SET_PROGRAM_TRACE_LEVEL((ThreadTraceLevel)(tmp));

       }
       else
       {
           cerr << "Error: invalid option: " << option << endl;
           exit(1);
       }
   }



   string dash_c_command_value;

   if (dash_c_option)
   {
       // rest of the command line, after the options is the command to execute

       if(!*argv)
       {
          cerr << "Error:  Missing the command for the -c options.\n";
          exit(1);
       }

       dash_c_command_value += *argv++;

       if(*argv)
       {
          cerr << "Error:  Too many arguments, Only one -c command may appear on the command line\n";
          exit(1);
       }

       launchFileReader = true;

   }
   else
   {
      // their should only be the name of the file to run

      if(*argv == 0)
      {
          launchConsoleReader = true;
      }
      else
      {
          commandString = *argv++;  // who knows what this option is.  probably a file name but we don't support that

      }

   }
                                                   // BOOL WINAPI IsDebuggerPresent(void);


   env.set("BRASH_THREAD_COUNT", lexical_cast<string>(threadCount) );


   BrashScriptCommandHandler handler(threadCount);  // 4 is the count of threads to start 

   brashGlobalScriptCommandHandler = &handler;

   if(launchFileReader)
   {
       if(dash_c_option)
       {
           stringstream ss(dash_c_command_value);

           handler.push( new ReadFileCommand(&ss, dash_c_command_value, &handler, env) );

           handler.waitForAllThreadsToTerminate();

           exit(handler.exitCode());

       }
       else
       {
           std::vector<std::string> args;

           args.push_back(scriptFileName);

           if(!commandString.empty())
           {
              // arg1 is already parsed by this point

              args.push_back(commandString);

              // now get the rest

              while(*argv)
              {
                 args.push_back(*argv++);
              }
           }

           handler.push( new ReadFileCommand(scriptFileName, &handler, env, args) );
       }
   }
   else
   if(launchConsoleReader)
   {
       handler.push( new ReadConsoleCommand(&handler, env) );
   }
   else
   {
      cerr << "Error:  brash does not process files on the command line unless you use the -script option\r\n";
      exit(1);
   }

   handler.waitForAllThreadsToTerminate();

   exit(handler.exitCode());

}
catch(std::exception const &e)
{
   cerr << "brash FATAL unhandled exception: " << e.what() << endl;
   exit(1);
}
} // main()




static void setBrashUserHome(StringExpr::ExecutionVariables &env)   // set variable BRASH_USER_HOME
{
    string HOMEVAR = env.get("HOME");

    string homedir;

    if(!HOMEVAR.empty())
    {
        FileName tmp(HOMEVAR);

        if(tmp.is_dir())
        {
           homedir = HOMEVAR;
        }
    }

    if(homedir.empty())
    {
#ifdef _MSC_VER
        string USERPROFILEVAR = env.get("USERPROFILE");

        FileName tmp(USERPROFILEVAR);

        if(tmp.is_dir())
        {
           homedir = USERPROFILEVAR;
        }
#endif
    }

    if(homedir.empty())
    {
       cerr << "brash: error:  unable to determine home directory\r\n" << endl;
#ifdef _MSC_VER
       homedir="c:\\";
#else
       homedir="c:\\";
#endif
    }

    env.set("BRASH_USER_HOME", homedir);

    env.set("BRASH_SHELL", interpreterExecutableName);
    env.set("BRASH_VERSION", BRASH_VERSION_STRING);

#ifdef _MSC_VER
    env.set("BRASH_PATH_SEPARATOR", ";");
    env.set("BRASH_DIR_SEPARATOR", "\\");
    env.set("BRASH_SYSROOT", "C:\\");
    env.set("BRASH_OS", "MSWindows");
    env.set("BRASH_USER", env.get("USERNAME"));
    env.set("BRASH_HOSTNAME", env.get("COMPUTERNAME"));
#else
    env.set("BRASH_PATH_SEPARATOR", ":");
    env.set("BRASH_DIR_SEPARATOR", "/");
    env.set("BRASH_SYSROOT", "");
    env.set("BRASH_OS", "Unix");

    string USERVAR = env.get("USER");

    if(USERVAR.empty())
       USERVAR = env.get("LOGNAME");

    evn.set("BRASH_USER", USERVAR);

    env.set("BRASH_HOSTNAME", env.get("HOSTNAME"));

#endif

    char buffer[2054];

    getcwd(buffer, sizeof(buffer));

    buffer[sizeof(buffer)-1] = 0;

    env.set("DIRSTACK",buffer); 

}


static std::string secondsUndefinedVariableHandler(cxxtls::StringExpr::ExecutionContext const* context)
{
   time_t now=0;

   time(&now);

   return lexical_cast<std::string>(now);

}

Mutex randomGuard;

unsigned int lfsr113_Bits (void)
{
   static unsigned int z1 = 12345, z2 = 12345, z3 = 12345, z4 = 12345;
   unsigned int b;
   b  = ((z1 << 6) ^ z1) >> 13;
   z1 = ((z1 & 4294967294U) << 18) ^ b;
   b  = ((z2 << 2) ^ z2) >> 27; 
   z2 = ((z2 & 4294967288U) << 2) ^ b;
   b  = ((z3 << 13) ^ z3) >> 21;
   z3 = ((z3 & 4294967280U) << 7) ^ b;
   b  = ((z4 << 3) ^ z4) >> 12;
   z4 = ((z4 & 4294967168U) << 13) ^ b;
   return (z1 ^ z2 ^ z3 ^ z4);
}

static std::string randomUndefinedVariableHandler(cxxtls::StringExpr::ExecutionContext const* context)
{
   ScopedMutexLock lock(randomGuard);

   return lexical_cast<std::string>(lfsr113_Bits());

}


static StringExpr::ExecutionContext::UndefinedVariablesHandlers undefinedVariablesMap;



static void fixLeadingTilde(std::string &value, cxxtls::StringExpr::ExecutionContext * context)
{
    if(!value.empty() && value[0] == '~')
    {
        if(value.size() == 1)
           value = context->find("BRASH_USER_HOME");
        else
        {
            char c= value[1];
            
            if(c == '/' || c == '\\')
            {
               value = context->find("BRASH_USER_HOME") + value.substr(1);
            }
        }
    }
}


static void SetStringExpressionUndefinedVariablesHandlers()
{

    undefinedVariablesMap.insert( StringExpr::ExecutionContext::UndefinedVariablesHandlers::value_type("SECONDS", &secondsUndefinedVariableHandler) );
    undefinedVariablesMap.insert( StringExpr::ExecutionContext::UndefinedVariablesHandlers::value_type("RANDOM",  &randomUndefinedVariableHandler) );


    StringExpr::ExecutionContext::registerUndefinedVariablesHandlers(&undefinedVariablesMap);

    StringExpression::registerPreGlobFilterFunction(fixLeadingTilde);

}

